/*
 * Copyright (c) 2008, CIPmap-authors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The names of CIPmap's contributors may not be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "cipmap.h"
#include "cm-util.h"
#include "cm-config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cairo.h>
#include <glib.h>
#include <libxml/tree.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>

static void testxml(void);
static gboolean run_connect_to_woist_server(gpointer data);

typedef struct CmPerson_ CmPerson;
typedef struct Workstation_ Workstation;
typedef struct Table_ Table;
typedef struct Map_ Map;
typedef struct CmInfoJob_ CmInfoJob;
typedef enum CmInfoJobType_ CmInfoJobType;

struct CmPerson_ {
    gchar *username;
    gchar *name;
    gchar *home;
    cairo_surface_t *img;
};

struct Workstation_ {
    gchar *name;
    gint x;
    gint y;
    GList *people;
};

struct Table_ {
    gint x;
    gint y;
    gint width;
    gint height;
};

struct Map_ {
    gint width;
    gint height;
    GList *workstations; /* workaround for 2.12: can't iterate over hash tables */
    GHashTable *workstation_table;
    GList *tables;
    gint gridcellsize;
    gint workstation_width;
    gint workstation_height;
    cairo_surface_t *workstation_img;
};

enum CmInfoJobType_ {
    CM_INFO_JOB_TYPE_INFO = 1,
    CM_INFO_JOB_TYPE_FACE
};

struct CmInfoJob_ {
    CmInfoJobType type;
    CmPerson *person;
};

static GList *people; /* workaround for 2.12: can't iterate over hash tables */
static GHashTable *people_table;
static Map *map;

static GHashTable *images;

static GIOChannel *logfile;
static GIOChannel *sock;
static gboolean draw_scheduled;
static GQueue *info_queue;

GQuark cm_xml_error_quark(void) {
    return g_quark_from_static_string("cm-xml-error");
}

void cm_log_init(void) {
    GError *error = NULL;
    gchar *file = g_strdup_printf("%s/cipmap.log", cm_config_get("wwwdir"));
    logfile = g_io_channel_new_file(file, "w", &error);
    if (logfile == NULL) {
        g_critical("couldn't open logfile: %s", error->message);
    }
}

/* TODO: add timestamp */
/* TODO: CM_ERROR prefixes msg with something like "ERR: " */
void cm_log_write(gint level, gchar *msg) {
    GError *error = NULL;
    
    fprintf(stderr, msg);
    
    if (level > CM_INFO || logfile == NULL) {
        return;
    }
    if ((g_io_channel_write_chars(logfile, msg, -1, NULL, &error)
         != G_IO_STATUS_NORMAL)
        || (g_io_channel_flush(logfile, &error) != G_IO_STATUS_NORMAL)) {
        g_critical("couldn't write to logfile: %s", error->message);
    }
    g_free(msg);
}

static cairo_surface_t *cm_images_get(const gchar *name) {
    return g_hash_table_lookup(images, name);
}

static CmPerson *cm_person_new(gchar *username, gchar *name,
                               cairo_surface_t *img) {
    CmPerson *person = g_new(CmPerson, 1);
    person->username = username;
    person->name = name;
    person->img = img;
    person->home = NULL;
    return person;
}

static void cm_person_free(CmPerson *person) {
    g_free(person);
}

static gchar *cm_person_to_string(CmPerson *person) {
    return g_strdup_printf("CmPerson[%s, %s, %s]", person->username,
                           person->name, person->home);
}

static Table *cm_table_new(gint x, gint y, gint width, gint height) {
    Table *table = g_new(Table, 1);
    table->x = x;
    table->y = y;
    table->width = width;
    table->height = height;
    return table;
}

static void cm_table_free(Table *table) {
    g_free(table);
}

static gchar *cm_table_to_string(Table *table) {
    return g_strdup_printf("Table[(%d, %d), %d, %d]", table->x, table->y,
                           table->width, table->height);
}

static Workstation *cm_workstation_new(gchar *name, gint x, gint y) {
    Workstation *ws = g_new(Workstation, 1);
    ws->name = name;
    ws->x = x;
    ws->y = y;
    ws->people = NULL;
    return ws;
}

static void cm_workstation_free(Workstation *ws) {
    g_free(ws);
}

static gchar *cm_workstation_to_string(Workstation *ws) {
    return g_strdup_printf("Workstation[%s, (%d, %d)]", ws->name, ws->x, ws->y);
}

static CmInfoJob *cm_info_job_new(CmInfoJobType type, CmPerson *person) {
    CmInfoJob *job = g_new(CmInfoJob, 1);
    job->type = type;
    job->person = person;
    return job;
}

static void cm_info_job_free(CmInfoJob *job) {
    g_free(job);
}

static void cm_draw_map(Map *map, cairo_surface_t *surface) {
    cairo_t *cr;
    gint x;
    gint y;
    GList *iter;
    Table *table;
    Workstation *ws = NULL;
    gfloat shadow_offset;
    gfloat ws_size;
    gfloat ws_offset;
    gfloat ws_scale;
    
    cr = cairo_create(surface);
    
    /* draw floor */
    cairo_set_line_width(cr, 1);
    for (y = 0; y < map->height; y++) {
        for (x = 0; x < map->width; x++) {
            cairo_rectangle(cr, x * map->gridcellsize, y * map->gridcellsize,
                            map->gridcellsize, map->gridcellsize);
            cairo_set_source_rgba(cr, 0.88, 0.88, 0.88, 1);
            cairo_fill_preserve(cr);
            cairo_set_source_rgba(cr, 0, 0, 0, 0.25);
            cairo_stroke(cr);
        }
    }
    
    /* draw table shadows */
    shadow_offset = 0.2 * map->gridcellsize;
    cairo_set_source_rgba(cr, 0, 0, 0, 0.25);
    for (iter = map->tables; iter != NULL; iter = iter->next) {
        table = iter->data;
        cairo_rectangle(cr, table->x * map->gridcellsize + shadow_offset,
                            table->y * map->gridcellsize + shadow_offset,
                            table->width * map->gridcellsize,
                            table->height * map->gridcellsize);
        cairo_fill(cr);
    }
    
    /* draw tables */
    cairo_set_line_width(cr, 2);
    for (iter = map->tables; iter != NULL; iter = iter->next) {
        table = iter->data;
        cairo_rectangle(cr, table->x * map->gridcellsize,
                            table->y * map->gridcellsize,
                            table->width * map->gridcellsize,
                            table->height * map->gridcellsize);
        cairo_set_source_rgba(cr, 0.80, 0.63, 0.20, 1);
        cairo_fill_preserve(cr);
        cairo_set_source_rgba(cr, 0.58, 0.47, 0.18, 1);
        cairo_stroke(cr);
    }
    
    /* draw workstations */
    cairo_select_font_face(cr, "Bitstream Vera Sans", CAIRO_FONT_SLANT_NORMAL,
                           CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(cr, 12);
    
    cairo_surface_t *workstation_img = cm_images_get("workstation");
    cairo_surface_t *person_img = cm_images_get("person");
    
    ws_size = 0.8 * map->gridcellsize;
    ws_scale = ws_size / cairo_image_surface_get_width(workstation_img);
    ws_offset = 0.1 * map->gridcellsize;
    
    gfloat person_offset_x = ws_offset;
    gfloat person_offset_y = ws_offset - 0.3 * ws_size;
    gfloat person_size = ws_size;
    gfloat person_scale = person_size / cairo_image_surface_get_height(person_img);
    cairo_surface_t *img = NULL;
    
    for (iter = map->workstations; iter != NULL; iter = iter->next) {
        ws = iter->data;
        
        cairo_translate(cr, ws->x * map->gridcellsize + ws_offset,
                            ws->y * map->gridcellsize + ws_offset);
        cairo_scale(cr, ws_scale, ws_scale);
        cairo_set_source_surface(cr, workstation_img, 0, 0);
        cairo_paint(cr);
        cairo_identity_matrix(cr);
        
        cairo_set_source_rgba(cr, 0, 0, 0, 1);
        cairo_move_to(cr, ws->x * map->gridcellsize + ws_offset, ws->y * map->gridcellsize + map->gridcellsize - ws_offset);
        cairo_show_text(cr, ws->name);
        
        GList *people_iter = ws->people;
        for (; people_iter != NULL; people_iter = people_iter->next) {
            CmPerson *person = people_iter->data;
            
            img = person->img;
            if (img == NULL) {
                img = person_img;
            } else {
                g_debug("custom image %s", person->username);
            }
            person_scale = person_size / cairo_image_surface_get_height(img);
            
            cairo_translate(cr, ws->x * map->gridcellsize + person_offset_x,
                                ws->y * map->gridcellsize + person_offset_y);
            cairo_scale(cr, person_scale, person_scale);
            cairo_set_source_surface(cr, img, 0, 0);
            cairo_paint(cr);
            cairo_identity_matrix(cr);
            
            cairo_set_source_rgba(cr, 0, 0, 0, 1);
            cairo_move_to(cr, ws->x * map->gridcellsize + person_offset_x,
                              ws->y * map->gridcellsize + person_offset_y + person_size);
            cairo_show_text(cr, person->username);
        }
    }
    
    /* draw people */
    /*gfloat person_offset_x = ws_offset;
    gfloat person_offset_y = ws_offset - 0.3 * ws_size;
    gfloat person_size = ws_size;
    gfloat person_scale = person_size / cairo_image_surface_get_width(person_img);
    cairo_surface_t *img = NULL;
    for (iter = people; iter != NULL; iter = iter->next) {
        person = iter->data;
        ws = person->ws;
        if (ws == NULL) {
            continue;
        }
        img = person->img;
        if (img == NULL) {
            img = person_img;
        }
        
        cairo_translate(cr, ws->x * map->gridcellsize + person_offset_x,
                            ws->y * map->gridcellsize + person_offset_y);
        cairo_scale(cr, person_scale, person_scale);
        cairo_set_source_surface(cr, img, 0, 0);
        cairo_paint(cr);
        cairo_identity_matrix(cr);
        
        cairo_set_source_rgba(cr, 0, 0, 0, 1);
        cairo_move_to(cr, ws->x * map->gridcellsize + person_offset_x,
                          ws->y * map->gridcellsize + person_offset_y + person_size);
        cairo_show_text(cr, person->username);
    }*/
}

static gboolean cm_create_png(gchar *pngfile) {
    cairo_surface_t *surface;
    
    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
        map->width * map->gridcellsize, map->height * map->gridcellsize);
    cm_draw_map(map, surface);
    
    if (cairo_surface_write_to_png(surface, pngfile) !=
        CAIRO_STATUS_SUCCESS) {
        fprintf(stderr, "failed to write PNG file\n");
        return FALSE;
    }
    
    cairo_surface_destroy(surface);
    return TRUE;
}

static void cm_create_html(void) {
    GError *error = NULL;
    gchar *templ = NULL;
    gchar *templ_file;
    
    templ_file = g_strdup_printf("%s/www/map.templ", RESDIR);
    
    if (!g_file_get_contents(templ_file, &templ, NULL, &error)) {
        fprintf(stderr, "failed to load template: %s\n", error->message);
        return;
    }
    
    /* creating webstuff... */
    
    gchar *mapname = "cip2";
    gchar *replaced = NULL;
    gchar *maphtml_file;
    gchar *xhtml;
    
    /* do for every map, and before that create $links */
    gchar *links = g_strdup_printf("<a href=\"%s.html\">%s</a>", mapname, mapname);
    
    xhtml = cm_strreplace(templ, "${map}", mapname);
    replaced = cm_strreplace(xhtml, "${links}", links);
    g_free(xhtml);
    xhtml = replaced;
    
    maphtml_file = g_strdup_printf("%s/%s.html", cm_config_get("wwwdir"), mapname);
    
    if(!g_file_set_contents(maphtml_file, xhtml, strlen(xhtml), &error)) {
        fprintf(stderr, "failed to write map-xhtml-file: %s\n", error->message);
        return;
    }
}

static gboolean cm_run_create_all(gpointer data) {
    cm_log(CM_INFO, "CREATING EVERYTHING");
    gchar *file = g_strdup_printf("%s/cip2.png", cm_config_get("wwwdir"));
    cm_create_png(file);
    g_free(file);
    cm_create_html();
    draw_scheduled = FALSE;
    return FALSE;
}

static void process_info_queue(void) {
    GError *error = NULL;
    
    GQueue *queue = info_queue;
    info_queue = g_queue_new();
    
    gboolean ssh_timeout = TRUE;
    
    CmInfoJob *job = g_queue_pop_head(queue);
    while (job != NULL) {
        CmPerson *person = job->person;

        gchar *out = NULL;
        gchar *err = NULL;
        gint exit_status = 0;
        gchar *cmdline = NULL;
        
        if (job->type == CM_INFO_JOB_TYPE_INFO) {
            //cmdline = g_strdup_printf("ssh cip \"getent passwd %s\"", job->person->username);
            cmdline = g_strdup_printf("ssh cip \"getent passwd %s\"", person->username);
            
            if (!g_spawn_command_line_sync(cmdline, &out, &err, &exit_status, &error)) {
                g_critical("ssh failed: %s", error->message);
                g_queue_push_tail(info_queue, job);
            } else {
                if (exit_status != 0) {
                    g_critical("no information for %s", job->person->username);
                    //g_queue_push_tail(info_queue, job);
                } else {
                    /* username:##username:x:x:name, office:home:shell */
                    gchar **tokens = g_strsplit(out, ":", 0);
                    person->name = g_strdup(tokens[4]);
                    person->home = g_strdup(tokens[5]);
                    g_strfreev(tokens);
                    CmInfoJob *job_face = cm_info_job_new(CM_INFO_JOB_TYPE_FACE, person);
                    g_queue_push_tail(info_queue, job_face);
                    
                    cm_info_job_free(job);
                    g_debug("info gathered %s", cm_person_to_string(person));
                }
            }
            
        } else if (job->type == CM_INFO_JOB_TYPE_FACE) {
            cmdline = g_strdup_printf("scp sisvpfal@faui05.informatik.uni-erlangen.de:%s/.face.png %s/faces/%s.png", person->home, RESDIR, person->username);
            g_debug(cmdline);
            
            if (!g_spawn_command_line_sync(cmdline, &out, &err, &exit_status, &error)) {
                g_critical("scp failed: %s", error->message);
                g_queue_push_tail(info_queue, job);
            } else {
                if (exit_status != 0) {
                    g_critical("scp failed: %d, %s", exit_status, err);
                    if (exit_status != 256) {
                        g_queue_push_tail(info_queue, job);
                    }
                } else {
                    cm_info_job_free(job);
                    g_debug("face copied");
                    
                    /* TDOO: check for errors */
                    gchar *file = g_strdup_printf("%s/faces/%s.png", RESDIR, person->username);
                    person->img = cairo_image_surface_create_from_png(file);
                    if (cairo_surface_status(person->img) != CAIRO_STATUS_SUCCESS) {
                        g_debug("failed to load default face image\n");
                        person->img = NULL;
                    } else {
                        g_debug("load face sucess");
                    }
                    g_free(file);
                }
            }
            
        }
        
        if (ssh_timeout) {
            job = g_queue_pop_head(queue);
            while (job != NULL) {
                g_queue_push_tail(info_queue, job);
                job = g_queue_pop_head(queue);
            }
            break;
        }
        
        job = g_queue_pop_head(queue);
    }
    
    g_queue_free(queue);
}

static gboolean run_process_info_queue(gpointer data) {
    process_info_queue();
    return TRUE;
}

static gboolean process_user_events(void) {
    GError *error = NULL;
    gchar *msg = NULL;
    
    gint status = g_io_channel_read_line(sock, &msg, NULL, NULL, &error);
    while (status == G_IO_STATUS_NORMAL) {
        
        /* parsing */
        /* action type workstation username\n */
        gchar *string = msg;

        gchar *delimiter = " \n";
        gchar *action = cm_strnext(&string, delimiter);
        cm_strnext(&string, delimiter); /* type */
        gchar *ws_name = cm_strnext(&string, delimiter);
        gchar *username = cm_strnext(&string, delimiter);
        
        if (action == NULL || ws_name == NULL || username == NULL) {
            g_warning("illegal message");
            status = g_io_channel_read_line(sock, &msg, NULL, NULL, &error);
            continue;
        }
        
        g_free(msg);
        
        if (g_str_equal(action, "reset")) {
            /* TODO: person_free */
            g_hash_table_remove_all(people_table);
            people = NULL;
            
        } else if (g_str_equal(action, "login") || g_str_equal(action, "logout")) {
            CmPerson *person = g_hash_table_lookup(people_table, username);
            if (person == NULL) {
                person = cm_person_new(username, "?", NULL);
                people = g_list_append(people, person);
                g_hash_table_insert(people_table, person->username, person);
                
                CmInfoJob *job = cm_info_job_new(CM_INFO_JOB_TYPE_INFO, person);
                g_queue_push_tail(info_queue, job);
                
                /* check for cached face image */
                gchar *file = g_strdup_printf("%s/faces/%s.png", RESDIR, person->username);
                person->img = cairo_image_surface_create_from_png(file);
                if (cairo_surface_status(person->img) != CAIRO_STATUS_SUCCESS) {
                    //g_debug("failed to load default face image\n");
                    person->img = NULL;
                } else {
                    g_debug("load face sucess");
                }
                g_free(file);
            }
            
            Workstation *ws = g_hash_table_lookup(map->workstation_table, ws_name);
            if (ws != NULL)  {
                if (g_str_equal(action, "login")) {
                    ws->people = g_list_append(ws->people, person);
                } else {
                    ws->people = g_list_remove(ws->people, person);
                }
            }
        }

        status = g_io_channel_read_line(sock, &msg, NULL, NULL, &error);
    }
    
    if (status == G_IO_STATUS_EOF) {
        cm_log(CM_INFO, "lost connection to Woist-Server\n");
        g_timeout_add(60 * 1000, (GSourceFunc) run_connect_to_woist_server, NULL);
        return FALSE;
        
    } else if (status == G_IO_STATUS_AGAIN) {
        if (!draw_scheduled) {
            draw_scheduled = TRUE;
            g_timeout_add(5 * 1000, (GSourceFunc) cm_run_create_all, NULL);
        }
        return TRUE;
        
    } else if (status == G_IO_STATUS_ERROR) {
        g_critical("couldn't read from stream: %s", error->message);
        return TRUE;
    }
    
    return TRUE;
}

static gboolean run_process_user_events(GIOChannel *sock,
                                        GIOCondition condition, gpointer data) {
    return process_user_events();
}

static gboolean connect_to_woist_server(void) {
    GError *error = NULL;
    cm_log(CM_INFO, "connecting to Woist-Server... ");
    
    gint fd = socket(PF_INET, SOCK_STREAM, 0);
    if (fd == -1) {
        g_critical("couldn't create socket: %s", strerror(errno));
        cm_log(CM_INFO, "failed\n");
        return FALSE;
    }
    
    struct addrinfo hints = {
        0, /* flags */
        AF_INET, /* family */
        SOCK_STREAM, /* socktype */
        0, /* protocol */
        0, /* addrlen */
        NULL, /* addr */
        NULL, /* canonname */
        NULL /* next */
    };
    struct addrinfo *address_info = NULL;
    gint e = 0;
    
    /* TODO: server address and port in configuration */
    //e = getaddrinfo("faui00c.informatik.uni-erlangen.de", "4712", NULL, &address_info);
    e = getaddrinfo("131.188.30.62", "54712", &hints, &address_info);
    if (e != 0) {
        g_critical("couldn't resolve hostname: %s", gai_strerror(e));
        cm_log(CM_INFO, "failed\n");
        return FALSE;
    }
    
    if (connect(fd, address_info->ai_addr, address_info->ai_addrlen) != 0) {
        g_debug("couldn't connect: %s", strerror(errno));
        cm_log(CM_INFO, "failed\n");
        return FALSE;
    }

    freeaddrinfo(address_info);
    
    sock = g_io_channel_unix_new(fd);
    if (g_io_channel_set_flags(sock, G_IO_FLAG_NONBLOCK, &error) != G_IO_STATUS_NORMAL) {
        g_critical("couldn't set non-blocking mode: %s", error->message);
        cm_log(CM_INFO, "failed\n");
        return FALSE;
    }
    
    g_io_add_watch(sock, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_PRI, run_process_user_events, NULL);
    
    cm_log(CM_INFO, "connected\n");
    return TRUE;
}

static gboolean run_connect_to_woist_server(gpointer data) {
    return !connect_to_woist_server();
}

static gboolean cm_setup_datadir(void) {
    GError *error = NULL;
    gboolean failed = FALSE;
    
    const gchar *datadir = cm_config_get("datadir");
    /*if (g_file_test(datadir, G_FILE_TEST_IS_DIR)) {
        cm_log(CM_INFO, "data directory is setup\n");
        return TRUE;
    }*/
    
    gchar *resdatadir = g_build_filename(cm_config_get("resdir"), "data", NULL);
    failed = !cm_copy(resdatadir, datadir, TRUE, &error);
    if (failed) {
        cm_log(CM_INFO, "failed to setup data directory: %s\n", error->message);
        g_error_free(error);
    } else {
        cm_log(CM_INFO, "successfully setup data directory\n");
    }
    
    g_free(resdatadir);
    return !failed;
}

static void init(void) {
    GError *error = NULL;
    draw_scheduled = FALSE;
    
    info_queue = g_queue_new();
    
    images = g_hash_table_new_full(g_str_hash, g_str_equal, 
        g_free, (GDestroyNotify) cairo_surface_destroy);
    
    gchar *imgdir = g_build_filename(cm_config_get("resdir"), "images", NULL);
    GList* items = cm_dir_get_items(imgdir, &error);
    if (error != NULL) {
        g_debug("failed to open imgdir; should exit program");
    }
    
    GList *iter = items;
    for (; iter != NULL; iter = iter->next) {
        gchar *imgfilename = iter->data;
        gchar *imgname = g_strdup(imgfilename);
        imgname[strlen(imgname) - 4] = '\0';
        gchar *imgfile = g_build_filename(imgdir, imgfilename, NULL);
        
        cairo_surface_t *img = cairo_image_surface_create_from_png(imgfile);
        if (cairo_surface_status(img) != CAIRO_STATUS_SUCCESS) {
            g_debug("failed to load image, should exit program");
        }
        g_hash_table_insert(images, imgname, img);
        
        g_free(imgfile);
    }
    cm_dir_item_list_free(items);
    g_free(imgdir);
    
    people_table = g_hash_table_new(g_str_hash, g_str_equal);
}

void quit(void) {
    cm_log(CM_INFO, "CIPmap exits\n");
    
    cm_config_clean();
}

int main(int argc, char **argv) {
    GError *error = NULL;
    
    cm_config_init();
    if (!cm_setup_datadir()) {
        quit();
    }
    
    gchar *file = g_strdup_printf("%s/config.xml", cm_config_get("datadir"));
    if (!cm_config_load_from_file(file, &error)) {
        cm_log(CM_INFO, "couldn't parse config-file: %s", error->message);
    }
    g_free(file);
    
    GMainLoop *mainloop = g_main_loop_new(NULL, FALSE);
    
    cm_log_init();
    cm_log(CM_INFO, "CIPmap started\n");
    
    init();
    
    cm_log(CM_INFO, "loading maps\n");
    testxml();
    
    /* Connect to Woist-Server. If it fails, retry every x seconds. */
    if (!connect_to_woist_server()) {
        g_timeout_add(60 * 1000, (GSourceFunc) run_connect_to_woist_server, NULL);
    }
    
    /* process the info-queue now and every x seconds */
    process_info_queue();
    g_timeout_add(10 * 1000, (GSourceFunc) run_process_info_queue, NULL);
    
    g_main_loop_run(mainloop);
    return EXIT_SUCCESS;
}

int x(int argc, char **argv)
{  
    /* read values from db... */
    
    /* faui00s:4712 */
    
    GError *error = NULL;
    gchar *cmdline;
    gchar *std_out = NULL;
    gchar *std_err = NULL;
    gint exit_status;
    
    //cmdline = "ssh cip \"cipmap/listusers\"";
    cmdline = "ssh cip \"cipmap/listusers 10\"";
    //cmdline = "ssh uni \"cipmap/listusers 2\"";
    
    if (!g_spawn_command_line_sync(cmdline, &std_out, &std_err, &exit_status, &error)) {
        fprintf(stderr, "failed to list users: %s\n", error->message);
        return EXIT_FAILURE;
    }

    gchar *string = std_out;
    gchar *delimiter = " \t\n";
    while (cm_strnext_peek(string, delimiter) != NULL) {
        gchar *username;
        gchar *ws_name_long;
        gchar *ws_name;
        gchar *next;
        gchar *from;

        username = cm_strnext(&string, delimiter);
        ws_name_long = cm_strnext(&string, delimiter);
        cm_strnext(&string, delimiter); /* month */
        cm_strnext(&string, delimiter); /* day */
        cm_strnext(&string, delimiter); /* time */
        
        /* could be login-time or from */
        next = cm_strnext(&string, delimiter);
        if (next[0] == '(') {
            from = next;
        } else {
            /* FIXME: could crash if string is < 1 */
            /* after a "login-time" there could be or could not be a "from" */
            if (string[1] == '\n') {
                from = "";
            } else {
                from = cm_strnext(&string, delimiter);
            }
        }
        
        if (username == NULL || ws_name_long == NULL || from == NULL) {
            fprintf(stderr, "parsing user list failed\n");
            break;
        }
        
        gchar *ws_name_long_string = ws_name_long;
        ws_name = cm_strnext(&ws_name_long_string, ".");
        ws_name_long_string = NULL;
        g_free(ws_name_long);
        
        if (ws_name == NULL) {
            fprintf(stderr, "parsing user list failed\n");
            break;
        }
        
        if (g_str_equal(from, "(:0.0)")) {
             /**NOTE: temp, only add users once */
            if (g_hash_table_lookup(people_table, username) == NULL) {
                CmPerson *person = cm_person_new(username, "?", NULL);
                //person->ws = g_hash_table_lookup(map->workstation_table, person->ws_name);
                people = g_list_append(people, person);
                g_hash_table_insert(people_table, person->username, person);
            }
        }
    }
    string = NULL;
    g_free(std_out);
    
    /*CmPerson *peter = cm_person_new("sipekran", "Peter Kranz", "faui00s", NULL);
    peter->ws = g_hash_table_lookup(map->workstation_table, peter->ws_name);
    people = g_list_append(people, peter);
    CmPerson *rainer = cm_person_new("sirrmuel", "Rainer Müller", "faui00r", NULL);
    rainer->ws = g_hash_table_lookup(map->workstation_table, rainer->ws_name);
    people = g_list_append(people, rainer);
    CmPerson *sven = cm_person_new("sisvpfal", "Sven Pfaller", "faui00d", NULL);
    sven->ws = g_hash_table_lookup(map->workstation_table, sven->ws_name);
    people = g_list_append(people, sven);*/
    
    cm_list_print(map->workstations, "workstations",
                  (CmToStringFunc) cm_workstation_to_string);
    cm_list_print(map->tables, "tables",
                  (CmToStringFunc) cm_table_to_string);
    cm_list_print(people, "people", (CmToStringFunc) cm_person_to_string);
    
    /* create image somewhere here */
    

    
    /*file = g_strdup_printf("%s/cip2.svg", WWWDIR);
    cm_create_svg(file);
    g_free(file);*/
    
    /* ... */
    

    
	printf("hello world again\n");
    
	return EXIT_SUCCESS;
}

/*static void test(void) {
    cairo_t *c;
    cairo_text_extents_t char_size;
    cairo_pattern_t *linear_gradient;
    cairo_pattern_t *radial_gradient;
    
    c = cairo_create(surface);
    
    cairo_select_font_face(c, "Bitstream Vera Sans", CAIRO_FONT_SLANT_NORMAL,
                           CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(c, 32);
    
    cairo_set_source_rgba(c, 1, 1, 0, 0.5);
    cairo_paint(c);
    
    linear_gradient = cairo_pattern_create_linear(0, 0, 240, 240);
    cairo_pattern_add_color_stop_rgba(linear_gradient, 0, 0, 1, 0, 0.5);
    cairo_pattern_add_color_stop_rgba(linear_gradient, 1, 1, 0, 0, 0.5);
    
    radial_gradient = cairo_pattern_create_radial(120, 120, 10, 120, 120, 100);
    cairo_pattern_add_color_stop_rgba(radial_gradient, 0, 0, 0, 0, 1);
    cairo_pattern_add_color_stop_rgba(radial_gradient, 0.95, 0, 0, 0, 1);
    cairo_pattern_add_color_stop_rgba(radial_gradient, 1, 0, 0, 0, 0);
    
    cairo_set_source(c, linear_gradient);
    cairo_mask(c, radial_gradient);
    
    cairo_rectangle(c, 10, 10, 30, 30);
    cairo_fill(c);
    
    cairo_set_source_rgb(c, 0, 0, 0);
    cairo_move_to(c, 10, 210);
    cairo_text_extents(c, "x", &char_size);
    cairo_show_text(c, "x");
    
    cairo_set_line_width(c, 1);
    cairo_rectangle(c, 10, 210 - char_size.height, char_size.width,
                    char_size.height);
    cairo_stroke(c);
    
    cairo_set_source_rgb(c, 1, 0.5, 0);
    cairo_set_line_width(c, 5);
    cairo_rectangle(c, 30, 100, 200, 100);
    cairo_stroke(c);
    
    cairo_translate(c, 130, 150);
    cairo_rotate(c, 5 * M_PI / 180);
    cairo_rectangle(c, -90, -40, 180, 80);
    cairo_fill(c);
    cairo_identity_matrix(c);
    
    cairo_set_source_rgb(c, 1, 0, 0);
    cairo_move_to(c, 10, 50);
    cairo_show_text(c, "Hello World!");
    
    cairo_set_source_rgba(c, 0, 0, 0, 0.5);
    cairo_set_line_width(c, 3);
    cairo_move_to(c, 40, 70);
    cairo_rel_line_to(c, 50, 20);
    cairo_rel_line_to(c, 50, -20);
    cairo_arc(c, 140, 100, 30, -90 * M_PI / 180, 90 * M_PI / 180);
    cairo_rel_curve_to(c, -20, -20, -40, 20, -60, 0);
    cairo_close_path(c);
    cairo_stroke(c);
    
    cairo_pattern_destroy(linear_gradient);
    cairo_pattern_destroy(radial_gradient);
    cairo_destroy(c);
}*/

static gboolean parse_xml_workstations(xmlDoc *doc, xmlNode *node) {
    xmlNode *child;
    gchar *value;
    gchar *name;
    gint x;
    gint y;
    
    for (child = node->children; child != NULL; child = child->next) {
        if (child->type == XML_COMMENT_NODE) {
            continue;
        }
        
        if (xmlStrcmp(child->name, (const xmlChar *) "workstation") == 0) {
            name = (gchar *) xmlGetProp(child, (const xmlChar *) "name");
            if (name == NULL) {
                fprintf(stderr, "error: workstation is missing name\n");
                return FALSE;
            }
            
            value = (gchar *) xmlGetProp(child, (const xmlChar *) "x");
            if (value == NULL) {
                fprintf(stderr, "error: workstation is missing x\n");
                return FALSE;
            }
            x = strtol(value, NULL, 10);
            
            value = (gchar *) xmlGetProp(child, (const xmlChar *) "y");
            if (value == NULL) {
                fprintf(stderr, "error: workstation is missing y\n");
                return FALSE;
            }
            y = strtol(value, NULL, 10);
            
            Workstation *ws = cm_workstation_new(name, x, y);
            map->workstations = g_list_append(map->workstations, ws);
            g_hash_table_insert(map->workstation_table, ws->name, ws);
            
        } else {
            fprintf(stderr, "error: xunknown element %s\n", child->name);
            return FALSE;
        }
    }
    
    return TRUE;
}

static gboolean parse_xml_decoration(xmlDoc *doc, xmlNode *node)
{
    xmlNode *child;
    gchar *value;
    gint x;
    gint y;
    gint width;
    gint height;
    
    for (child = node->children; child != NULL; child = child->next) {
        if (child->type == XML_COMMENT_NODE) {
            continue;
        }
        
        if (xmlStrcmp(child->name, (const xmlChar *) "table") == 0) {
            value = (gchar *) xmlGetProp(child, (const xmlChar *) "x");
            if (value == NULL) {
                fprintf(stderr, "error: table is missing x\n");
                return FALSE;
            }
            x = strtol(value, NULL, 10);
            
            value = (gchar *) xmlGetProp(child, (const xmlChar *) "y");
            if (value == NULL) {
                fprintf(stderr, "error: table is missing y\n");
                return FALSE;
            }
            y = strtol(value, NULL, 10);
            
            value = (gchar *) xmlGetProp(child, (const xmlChar *) "width");
            if (value == NULL) {
                fprintf(stderr, "error: table is missing width\n");
                return FALSE;
            }
            width = strtol(value, NULL, 10);
            
            value = (gchar *) xmlGetProp(child, (const xmlChar *) "height");
            if (value == NULL) {
                fprintf(stderr, "error: table is missing height\n");
                return FALSE;
            }
            height = strtol(value, NULL, 10);
            
            Table *table = cm_table_new(x, y, width, height);
            map->tables = g_list_append(map->tables, table);
            
        } else {
            fprintf(stderr, "error: unknown element %s\n", child->name);
            return FALSE;
        }
    }
    
    return TRUE;
}

static gboolean parse_xml_map(xmlDoc *doc, xmlNode *node) {
    /* TODO: also return GError */
    
    map = g_new(Map, 1);
    map->width = 10;
    map->height = 10;
    map->workstations = NULL;
    map->workstation_table = g_hash_table_new(g_str_hash, g_str_equal);
    map->tables = NULL;
    map->gridcellsize = 50;
    map->workstation_width = 1;
    map->workstation_height = 1;
    map->workstation_img = NULL;
    
    xmlNode *child;
    xmlChar *value;
    
    for (child = node->children; child != NULL; child = child->next) {
        g_debug((gchar *) child->name);
        
        if (child->type == XML_COMMENT_NODE) {
            continue;
        }
        
        if (xmlStrcmp(child->name, (const xmlChar *) "gridcellsize") == 0) {
            gint gridcellsize;
            value = xmlNodeListGetString(doc, child->children, 1);
            if (value == NULL) {
                fprintf(stderr, "error: content missing\n");
                return FALSE;
            }
            gridcellsize = strtol((gchar *) value, NULL, 10);
            xmlFree(value);
            if (gridcellsize <= 0) {
                fprintf(stderr, "error: illegal value for gridcellsize\n");
                return FALSE;
            }
            map->gridcellsize = gridcellsize;
        
        } else if (xmlStrcmp(child->name, (const xmlChar *) "size") == 0) {
            gint width;
            gint height;
            value = xmlGetProp(child, (const xmlChar *) "width");
            if (value == NULL) {
                fprintf(stderr, "error: size is missing width\n");
                return FALSE;
            }
            width = strtol((gchar *) value, NULL, 10);
            xmlFree(value);
            if (width <= 0) {
                fprintf(stderr, "error: illegal value for width\n");
                return FALSE;
            }
            value = xmlGetProp(child, (const xmlChar *) "height");
            if (value == NULL) {
                fprintf(stderr, "error: size is missing height\n");
                return FALSE;
            }
            height = strtol((gchar *) value, NULL, 10);
            xmlFree(value);
            if (height <= 0) {
                fprintf(stderr, "error: illegal value for height\n");
                return FALSE;
            }
            map->width = width;
            map->height = height;
                
        } else if (xmlStrcmp(child->name, (const xmlChar *) "workstations") == 0) {
            parse_xml_workstations(doc, child);
        } else if (xmlStrcmp(child->name, (const xmlChar *) "decoration")
                   == 0) {
            parse_xml_decoration(doc, child);
        } else {
            fprintf(stderr, "error: unknown element %s\n", child->name);
            return FALSE;
        }
    }
    
    return TRUE;
}

static void testxml(void) {
    gchar *xmlfile;
    xmlDoc *doc;
    
    /* TODO: multimap */
    xmlfile = g_build_filename(cm_config_get("datadir"), "maps", "example.xml", NULL);
    doc = xmlReadFile(xmlfile, NULL, XML_PARSE_NOBLANKS);
    if (doc == NULL) {
        fprintf(stderr, "failed to parse XML file\n");
        return;
    }
    
    xmlNode *node;
    
    node = xmlDocGetRootElement(doc);
    
    if (node == NULL) {
        fprintf(stderr, "error: empty XML file\n");
        return;
    }
    if (xmlStrcmp(node->name, (const xmlChar *) "map") != 0) {
        g_print("%s\n", node->name);
        fprintf(stderr, "error: missing <map>\n");
        return;
    }
    
    parse_xml_map(doc, node);
    
    g_free(xmlfile);
}
