/*
 * Copyright (c) 2008, CIPmap-authors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The names of CIPmap's contributors may not be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "cm-config.h"
#include "cm-util.h"
#include "cipmap.h"
#include "config.h"

static GHashTable *config;

static void cm_config_parse_config(xmlDoc *doc, xmlNode *config_node);

void cm_config_init(void) {
    gchar *resdir = cm_strreplace(RESDIR, "HOME", g_get_home_dir());
    gchar *datadir = cm_strreplace(DATADIR, "HOME", g_get_home_dir());
    
    config = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, g_free);
    g_hash_table_insert(config, "resdir", resdir);
    g_hash_table_insert(config, "datadir", datadir);
    
    /* defaults */
    g_hash_table_insert(config, "woistserver_host", g_strdup("localhost"));
    g_hash_table_insert(config, "woistserver_port", g_strdup("4712"));
    g_hash_table_insert(config, "wwwdir", g_strdup("www"));
}

void cm_config_clean(void) {
    g_hash_table_destroy(config);
}

const gchar *cm_config_get(const gchar *name) {
    g_return_val_if_fail(name != NULL, NULL);
    return g_hash_table_lookup(config, name);
}

gboolean cm_config_load_from_file(const gchar *file, GError **error) {
    g_return_val_if_fail(file != NULL, FALSE);
    g_return_val_if_fail(error == NULL || *error == NULL, FALSE);

    /* TODO: supress stderr-output from xmlReadFile */
    xmlDoc *doc = xmlReadFile(file, NULL, XML_PARSE_NOBLANKS | XML_PARSE_NOERROR
                              | XML_PARSE_NOWARNING);
    if (doc == NULL) {
        if (error != NULL) {
            xmlError *e = xmlGetLastError();
            *error = g_error_new(CM_XML_ERROR, CM_XML_ERROR_TREE, e->message);
        }
        return FALSE;
    }
    
    xmlNode *root = xmlDocGetRootElement(doc);

    if (root == NULL) {
        if (error != NULL) {
            *error = g_error_new(CM_XML_ERROR, CM_XML_ERROR_PARSE,
                                 "config-file is empty");
        }
        xmlFreeDoc(doc);
        return FALSE;
    }
    
    if (xmlStrcmp(root->name, (const xmlChar *) "config") != 0) {
        if (error != NULL) {
            *error = g_error_new(CM_XML_ERROR, CM_XML_ERROR_PARSE,
                                 "file has wrong type (\"%s\")", root->name);
        }
        xmlFreeDoc(doc);
        return FALSE;
    }
    
    cm_config_parse_config(doc, root);
    
    xmlFreeDoc(doc);
    return TRUE;
}

/* does relaxed parsing */
static void cm_config_parse_config(xmlDoc *doc, xmlNode *config_node) {
    g_return_if_fail(config_node != NULL);
    g_return_if_fail(xmlStrcmp(config_node->name, (const xmlChar *) "config")
                     == 0);
    
    xmlNode *node = config_node->children;
    for (; node != NULL; node = node->next) {
        if (node->type == XML_COMMENT_NODE) {
            continue;
        }
        
        if (xmlStrcmp(node->name, (const xmlChar *) "woistserver") == 0) {
            xmlChar *host = xmlGetProp(node, (const xmlChar *) "host");
            xmlChar *port = xmlGetProp(node, (const xmlChar *) "port");
            if (host != NULL) {
                g_hash_table_insert(config, "woistserver_host", host);
            } else {
                cm_log(CM_DETAIL, "parsing config: woistserver_host missing\n");
            }
            if (port != NULL) {
                g_hash_table_insert(config, "woistserver_port", port);
            } else {
                cm_log(CM_DETAIL, "parsing config: woistserver_port missing\n");
            }
            
        } else if (xmlStrcmp(node->name, (const xmlChar *) "wwwdir") == 0) {
            xmlChar *wwwdir = xmlNodeListGetString(doc, node->children, 1);
            if (wwwdir != NULL) {
                g_hash_table_insert(config, "wwwdir", wwwdir);
            } else {
                cm_log(CM_DETAIL, "parsing config: wwwdir value missing\n");
            }
            
        } else {
            cm_log(CM_DETAIL, "parsing config: unknown element\n");
        }
    }
}

/* TODO: call xmlCleanupParser somewhere */
