/*
 * Copyright (c) 2008, CIPmap-authors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The names of CIPmap's contributors may not be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "cm-util.h"

#include <glib/gstdio.h>
#include <errno.h>

void cm_list_print(GList *list, const gchar *title,
                        CmToStringFunc to_string) {
    g_return_if_fail (title != NULL);
  
    g_print ("%s (%d):\n", title, g_list_length(list));
    for (; list != NULL; list = list->next) {
        if (to_string == NULL) {
          g_print(" * %s\n", (gchar *) list->data);
        } else {
          g_print(" * %s\n", to_string(list->data));
        }
    }
}

/*void cm_hash_table_print(GHashTable *table, const gchar *title,
    CmToStringFunc key_to_string, CmToStringFunc value_to_string) {
    GHashTableIter iter;
    gpointer key = NULL;
    gpointer value = NULL;
    
    g_hash_table_iter_init(&iter, table);
    while (g_hash_table_iter_next(&iter, &key, &value)) {
        if (key_to_string == NULL) {
            g_print(" * (%s", (gchar *) key);
        } else {
            g_print(" * (%s", key_to_string(key));
        }
        if (value_to_string == NULL) {
            g_print(", %s)\n", (gchar *) value);
        } else {
            g_print(", %s)\n", value_to_string(value));
        }
    }
}*/

gchar *cm_strreplace(const gchar *string, const gchar *victim, const gchar *replacement) {
    g_return_val_if_fail(string != NULL, NULL);
    g_return_val_if_fail(victim != NULL, NULL);
    g_return_val_if_fail(replacement != NULL, NULL);
    
    gchar **tokens;
    gchar *token;
    gchar **iter;
    GString *buffer;
    gchar *result;
    
    tokens = g_strsplit(string, victim, 0);
    iter = tokens;
    token = *iter;
    buffer = g_string_new(token);
    iter++;
        
    for (; *iter != NULL; iter++) {
        g_string_append(buffer, replacement);
        token = *iter;
        g_string_append(buffer, token);
    }
    
    result = buffer->str;
    g_string_free(buffer, FALSE);
    g_free(tokens);
    return result;
}

/* TODO: make the next methods work with UTF8 */
/* \0 must not be in delimiters */

gchar *cm_strnext(gchar **string, gchar *delimiters) {
    return cm_strnext_impl(string, delimiters, FALSE);
}

gchar *cm_strnext_peek(gchar *string, gchar *delimiters) {
    return cm_strnext_impl(&string, delimiters, TRUE);
}

gchar *cm_strnext_impl(gchar **string, gchar *delimiters, gboolean peek) {
    gchar *p = *string;
    
    /* Skip leading delimiters. If the end of the string is reached, there is
     * no next token. */
    while (cm_char_in_set(*p, delimiters)) {
        p++;
    }
    if (*p == '\0') {
        if (!peek) {
            *string = p;
        }
        return NULL;
    }
    gchar *start = p;
    
    /* consume next token (until a delimiter or the end of the string is reached) */
    gint len = 0;
    while (*p != '\0' && !cm_char_in_set(*p, delimiters)) {
        p++;
        len++;
    }
    if (!peek) {
        *string = p;
    }
    return g_strndup(start, len);
}

gboolean cm_char_in_set(gchar c, gchar *set) {
    while (*set != '\0') {
        if (c == *set) {
            return TRUE;
        }
        set++;
    }
    return FALSE;
}

gboolean cm_file_copy(const gchar *srcfile, const gchar *destfile, GError **error) {
    g_return_val_if_fail(srcfile != NULL, FALSE);
    g_return_val_if_fail(destfile != NULL, FALSE);
    g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
    gboolean failed = FALSE;
    
    GIOChannel* src = g_io_channel_new_file(srcfile, "r", error);
    if (src == NULL) {
        return FALSE;
    }
    GIOChannel* dest = g_io_channel_new_file(destfile, "w", error);
    if (dest == NULL) {
        g_io_channel_unref(src);
        return FALSE;
    }
    
    gchar *buffer = g_new(gchar, CM_BUFFERSIZE);
    gsize read = 0;
    
    GIOStatus status = g_io_channel_read_chars(src, buffer, CM_BUFFERSIZE, &read, error);
    while (status != G_IO_STATUS_EOF) {
        if (status == G_IO_STATUS_ERROR) {
            failed = TRUE;
            break;
        }
        g_assert(status != G_IO_STATUS_AGAIN);
        
        status = g_io_channel_write_chars(dest, buffer, read, NULL, error);
        if (status == G_IO_STATUS_ERROR) {
            failed = TRUE;
            break;
        }
        g_assert(status != G_IO_STATUS_EOF && status != G_IO_STATUS_AGAIN);
        
        status = g_io_channel_read_chars(src, buffer, CM_BUFFERSIZE, &read, error);
    }
    
    g_free(buffer);
    g_io_channel_unref(src);
    g_io_channel_unref(dest);
    return !failed;
}

/* does not copy if file exists */
gboolean cm_copy(const gchar *srcpath, const gchar *destpath, gboolean recursive, GError **error) {
    g_return_val_if_fail(srcpath != NULL, FALSE);
    g_return_val_if_fail(destpath != NULL, FALSE);
    g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
    gboolean failed = FALSE;
    
    /*g_debug("copy \"%s\" to \"%s\" (recursively: %d)", srcpath, destpath, recursive);*/

    /* copy a file */
    if (g_file_test(srcpath, G_FILE_TEST_IS_REGULAR)) {
        
        /* copy file to directory: same as copy file to file (with destination
         * being directory/filename) */
        gchar *destfile = g_strdup(destpath);
        if (g_file_test(destpath, G_FILE_TEST_IS_DIR)) {
            gchar *srcfilename = g_path_get_basename(srcpath);
            destfile = g_build_filename(destpath, srcfilename, NULL);
            g_free(srcfilename);
        }
        
        /* copy file to file */
        failed = !cm_file_copy(srcpath, destfile, error);
        
        g_free(destfile);
        return !failed;
    }
    
    /* copy a directory (recursively) */
    if (g_file_test(srcpath, G_FILE_TEST_IS_DIR)) {
        if (!recursive) {
            if (error != NULL) {
                *error = g_error_new(G_FILE_ERROR, G_FILE_ERROR_ISDIR, "source \"%s\" is a directory", srcpath);
            }
            return FALSE;
        }
        
        /* copy directory to directory: if the destination directory doesn't
         * exist, create it.
         * copy directory to file: error */
        if (!g_file_test(destpath, G_FILE_TEST_IS_DIR)) {
            if (g_mkdir(destpath, 0777) != 0) {
                if (error != NULL) {
                    *error = g_error_new(G_FILE_ERROR, g_file_error_from_errno(errno), g_strerror(errno));
                }
                return FALSE;
            }
        }

        GError *tmperror = NULL;
        GList *items = cm_dir_get_items(srcpath, &tmperror);
        if (tmperror != NULL) {
            g_propagate_error(error, tmperror);
            return FALSE;
        }
        
        GList *iter = items;
        for (; iter != NULL && !failed; iter = iter->next) {
            gchar *srcitem = g_build_filename(srcpath, iter->data, NULL);
            gchar *destitem = g_build_filename(destpath, iter->data, NULL);

            failed = !cm_copy(srcitem, destitem, recursive, error);

            g_free(srcitem);
            g_free(destitem);
        }
    
        cm_dir_item_list_free(items);
        return !failed;
    }
    
    /* the source file or directory does not exist */
    if (error != NULL) {
        *error = g_error_new(G_FILE_ERROR, G_FILE_ERROR_NOENT, "no file or directory \"%s\"", srcpath);
    }
    return FALSE;
}

GList *cm_dir_get_items(const gchar *path, GError **error) {
    GList *items = NULL;
    GDir *dir = g_dir_open(path, 0, error);
    if (dir == NULL) {
        return NULL;
    }
    
    gchar *item = g_strdup(g_dir_read_name(dir));
    while (item != NULL) {
        items = g_list_append(items, item);
        item = g_strdup(g_dir_read_name(dir));
    }
    
    g_dir_close(dir);
    return items;
}

void cm_dir_item_list_free(GList *dir_item_list) {
    GList *iter = dir_item_list;
    for (; iter != NULL; iter = iter->next) {
        g_free(iter->data);
    }
    g_list_free(dir_item_list);
}
